from conf import*       # fiber specific modules ( symbol, token ) and objects ( prompt, pprint, unparse ...)
import EasyExtend       # EasyExtend package
import parsetable       # contains fiber specific DFA tables used by the parser
from EasyExtend.csttools import*   # tools used to manipulate CSTs
from EasyExtend.eetransformer import Transformer, transform
from EasyExtend.util.path import path
import copy
import sets

def join_grammars(g_cst1, g_cst2):
    g_cst = load_reference_grammar()
    erase_line_numbers(g_cst1)
    erase_line_numbers(g_cst2)
    erase_line_numbers(g_cst)
    rule_set0 = find_all(g_cst, symbol.RULE)
    rule_set1 = find_all(g_cst1, symbol.RULE)
    rule_set2 = find_all(g_cst2, symbol.RULE)
    d_cst1 = {}
    def delta_rules(rule_set):
        delta_rule_set = []
        for k in range(len(rule_set)):
            name = find_node(rule_set[k], token.NAME)[1]
            if k<len(rule_set0):
                if name == find_node(rule_set0[k], token.NAME)[1]:
                    if check_rule_equivalence(rule_set[k], rule_set0[k]) == False:
                        delta_rule_set.append(name)
            else:
                delta_rule_set.append(name)
        return delta_rule_set
    delta_set1 = sets.Set(delta_rules(rule_set1))
    delta_set2 = sets.Set(delta_rules(rule_set2))
    print "delta_set1:"
    print "----------"
    for r in delta_set1:
        print r
    print
    print "delta_set2:"
    print "----------"
    for r in delta_set2:
        print r
    print
    for_fusion = delta_set2.intersection(delta_set1)
    print "DEBUG", for_fusion
    for_replacement1 = delta_set1 - for_fusion
    for_replacement2 = delta_set2 - for_fusion
    res_cst = [symbol.file_input]
    for k in range(len(rule_set0)):
        name = find_node(rule_set0[k], token.NAME)[1]
        if name in for_replacement1:
            res_cst.append(rule_set1[k])
            for_replacement1.difference_update(sets.Set([name]))
        elif name in for_replacement2:
            res_cst.append(rule_set2[k])
            for_replacement2.difference_update(sets.Set([name]))
        elif name in for_fusion:
            res_cst.append(fusion_rules(rule_set1[k], rule_set2[k]))
        else:
            res_cst.append(rule_set0[k])
    m = k

    def append_rules(for_append, rule_set):
        for k in range(m,len(rule_set)):
            name = find_node(rule_set[k], token.NAME)[1]
            if name in for_append:
                res_cst.append(rule_set[k])
    append_rules(for_replacement1, rule_set1)
    append_rules(for_replacement2, rule_set2)
    return res_cst

def RULE(*args):
    "RULE: NAME ':' RHS NEWLINE"
    assert len(args) == 2
    nodelist = [symbol.RULE]
    assert args[0][0] == token.NAME
    nodelist.append(args[0])
    assert args[1][0] == symbol.RHS
    nodelist.append(Symbol(token.COLON,':'))
    nodelist.append(args[1])
    nodelist.append(NEWLINE())
    return nodelist

def RHS(*args):
    "RHS: ALT ( '|' ALT )*"
    '''
    _grp = False
    if len(args)>1:
        _args = []
        for arg in args:
            if is_group(arg):
                _grp = True
                fl_args = flatten(arg)
                for a in fl_args:
                    if a not in _args:
                        _args.append(a)
            elif arg not in _args:
                _args.append(arg)
        args = _args
    '''
    assert len(args)>0
    nodelist = [symbol.RHS]
    for arg in args[:-1]:
        assert arg[0] == symbol.ALT
        nodelist.append(arg)
        nodelist.append(Symbol(token.VBAR, '|'))
    else:
        assert args[-1][0] == symbol.ALT
        nodelist.append(args[-1])
    return nodelist


def ALT(*args):
    "ALT: ITEM+"
    nodelist = [symbol.ALT]
    assert len(args)>0
    for arg in args:
        assert arg[0] == symbol.ITEM
        nodelist.append(arg)
    return nodelist

def ITEM(*args):
    "ITEM: '[' RHS ']' | ATOM [ '*' | '+' ]"
    nodelist = [symbol.ITEM]
    if args[0] == '[':
        nodelist.append(Symbol(token.LSQB, '['))
        assert args[1][0] == symbol.RHS
        nodelist.append(args[1])
        assert args[2] == ']'
        nodelist.append(Symbol(token.RSQB, ']'))
        return nodelist
    else:
        assert args[0][0] == symbol.ATOM
        nodelist.append(args[0])
        if len(args) == 2:
            if args[1] == '*':
                nodelist.append(Symbol(token.STAR, '*'))
            elif args[1] == '+':
                nodelist.append(Symbol(token.PLUS, '+'))
            else:
                raise ValueError("Unexpected symbol %s"%args[1])
        return nodelist

def ATOM(*args):
    "ATOM: '(' RHS ')' | NAME | STRING"
    nodelist = [symbol.ATOM]
    if args[0] == '(':
        nodelist.append(Symbol(token.LPAR, '('))
        assert args[1][0] == symbol.RHS
        nodelist.append(args[1])
        assert args[2] == ')'
        nodelist.append(Symbol(token.RPAR, ')'))
        return nodelist
    elif args[0][0] in (token.NAME, token.STRING):
        nodelist.append(args[0])
    else:
        raise ValueError("Unexpected symbol %s"%args[0])
    return nodelist

def erase_line_numbers(n):
    for item in n:
        if isinstance(item,list):
            if isinstance(item[-1],int):
                item[-1] = 0
            erase_line_numbers(item)

def check_rule_equivalence(r1, r2):
    if r1 == r2:
        return True
    # check for rule permutation by means of ALT clauses
    alt_rules1 = sorted(find_all(r1, symbol.ALT))
    alt_rules2 = sorted(find_all(r2, symbol.ALT))
    return alt_rules1 == alt_rules2

def is_group(node):
    _item = find_node(node, symbol.ITEM)
    if _item[1][0] == symbol.ATOM:
        _atom = find_node(_item, symbol.ATOM)
        if _atom[1][0] == token.LPAR:
            return True
    return False

def make_group(rhs):
    return RHS(ALT(ITEM(ATOM(*('(',rhs,')')))))


def flatten_group(node):
    '''
    (A|B|....) -> A,B,C,...
    '''
    if is_group(node):
        _alt = find_all(find_node(node, symbol.ATOM), symbol.ALT, level = 2)
        if len(_alt)>1:
            return _alt
    return [node]

def fusion_alt(r1, r2):
    selection = []
    flat_r1 = sum([flatten(r) for r in r1],[])
    for r in r2:
        if r not in flat_r1:
            selection.append(r)
    return r1, selection

def fusion_rules(r1, r2):
    name = find_node(r1, token.NAME)
    r1_alt = find_all(r1, symbol.ALT, level = 2)
    r2_alt = find_all(r2, symbol.ALT, level = 2)
    selection = []
    '''
    R1: A1 | B1 | ...
    R2: A2 | B2 | ...
    -----------
    R : A1 | A2 | B1 | B2 | ... | Ai1 | ^Ai2 | ... , with Ai1 in R1 and R2
    '''
    r1_alt, selection = fusion_alt(r1_alt, r2_alt)
    r1_alt, selection = fusion_left_items(r1_alt, selection)
    r1_alt, selection = fusion_right_items(r1_alt, selection)
    return RULE(name, RHS(*(r1_alt+selection)))

def fusion_left_items(r1, r2):
    '''
    R1: A C
    R2: B C
    -------
    (A|B) C
    '''
    r1_last_item = [alt[-1] for alt in r1]
    r2_last_item = [alt[-1] for alt in r2]
    for i, last in enumerate(r2_last_item):
        if last in r1_last_item:
            idx = r1_last_item.index(last)
            r1_alt = r1[idx]
            k = len(r1_alt)-1
            fin = []
            for j, item in enumerate(r2[i][::-1]):
                if item == r1_alt[k-j]:
                    fin.insert(0,item)
                else:
                    break
            pre1 = r1_alt[1:k+1-len(fin)]
            pre2 = r2[i][1:len(r2[i])-len(fin)]
            A1 = ALT(*pre1)
            A2 = ALT(*pre2)
            a = ITEM(ATOM('(',RHS(A1,A2),')'))
            r2[i] = ALT(*([a]+fin))
            del r1[idx]
            break
    return r1, r2

def fusion_right_items(r1, r2):
    '''
    R1: A B
    R2: A C
    -------
    A (B | C)
    '''
    r1_first_item = [item[1] for item in r1]
    r2_first_item = [item[1] for item in r2]
    for i, first in enumerate(r2_first_item):
        if first in r1_first_item:
            idx = r1_first_item.index(first)
            r1_alt = r1[idx]
            k = len(r1_alt)-1
            begin = []
            for j, item in enumerate(r2[i][1:]):
                if item == r1_alt[1:][j]:
                    begin.append(item)
                else:
                    break
            post1 = r1_alt[1+len(begin):]
            post2 = r2[i][1+len(begin):]
            A1 = ALT(*post1)
            A2 = ALT(*post2)
            z = ITEM(ATOM('(',RHS(A1,A2),')'))
            r2[i] = ALT(*(begin+[z]))
            del r1[idx]
            break
    return r1, r2




__publish__ = []

def load_fiber_grammar(F):
    import EasyExtend.fibers.grammar_fiber.fiber as grammar_fiber
    source = path(F.__file__).dirname().joinpath("Grammar").open().read()
    import EasyExtend.eecompiler as eecompiler
    eec = eecompiler.EECompiler(grammar_fiber)
    return eec.eeparse_suite(source)

def load_reference_grammar():
    import EasyExtend.fibers.grammar_fiber.fiber as grammar_fiber
    source = path(fiber.__file__).dirname().joinpath("GrammarRef").open().read()
    import EasyExtend.eecompiler as eecompiler
    eec = eecompiler.EECompiler(grammar_fiber)
    return eec.eeparse_suite(source)

def parse_rule(rule_expr):
    import fiber
    import EasyExtend.eecompiler as eecompiler
    eec = eecompiler.EECompiler(fiber)
    return find_node(eec.eeparse_suite(rule_expr), symbol.RULE)


if __name__ == '__main__':
    def test1():
        r1 = parse_rule("R: C | B")
        r2 = parse_rule("R: A | C")
        r3 = parse_rule("R: C | B | A")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: A | B")
        r2 = parse_rule("R: A | C")
        r3 = parse_rule("R: A | B | C")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: (A | B)")
        r2 = parse_rule("R: A")
        r3 = parse_rule("R: (A | B)")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)

    #test1()

    def test2():
        r1 = parse_rule("R: A C")
        r2 = parse_rule("R: B C")
        r3 = parse_rule("R: (A | B ) C")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: A C D")
        r2 = parse_rule("R: B C D")
        r3 = parse_rule("R: (A | B ) C D")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: A D")
        r2 = parse_rule("R: B C D")
        r3 = parse_rule("R: (A | B C ) D")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: A C D")
        r2 = parse_rule("R: B D")
        r3 = parse_rule("R: (A C | B ) D")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
    #test2()

    def test3():
        r1 = parse_rule("R: A B")
        r2 = parse_rule("R: A C")
        r3 = parse_rule("R: A( B | C )")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: A B D")
        r2 = parse_rule("R: A B C")
        r3 = parse_rule("R: A B ( D | C )")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)
        r1 = parse_rule("R: 'A' D E")
        r2 = parse_rule("R: 'A' C")
        r3 = parse_rule("R: 'A' ( D E | C )")
        assert unparse(fusion_rules(r1, r2)) == unparse(r3)

    #test3()
    def test4():
        r1 = parse_rule("R: (A|C) B")
        r2 = parse_rule("R: C B")
        assert unparse(fusion_rules(r1, r2)) == unparse(r1)
    #test4()

    def test5():
        r1 = parse_rule("R: ('+'|'~') factor | power")
        r2 = parse_rule("R: ('+'|'~'|'++') factor | power")
        print unparse(fusion_rules(r1, r2))

    # test5()
