import re
re_fiber = None

class regex(object):
    def __init__(self, regex_obj):
        if isinstance(regex_obj, basestring):
            self.obj = text(regex_obj)
        else:
            self.obj = regex_obj

    @classmethod
    def from_re(self, regexp):
        '''
        Parse regular expression string into regex object.
        '''
        global re_fiber
        if re_fiber is None:
            import EasyExtend
            import EasyExtend.eecommon
            import EasyExtend.fibers.re_fiber.fiber as re_fiber
            EasyExtend.eecommon.init_fiber(re_fiber)
        return eval(re_fiber.compile_re(regexp))

    @classmethod
    def _toregex(cls, obj):
        if isinstance(obj, regex):
            return obj
        else:
            return text(obj)

    def _get_some(self):
        return repeat(self, domain= slice(0,None,None), greedy = False)

    def _get_many(self):
        return repeat(self, domain= slice(1,None,None), greedy = True)

    def _get_many(self):
        return repeat(self, domain= slice(1,None,None), greedy = True)


    some = property(_get_some)
    many = property(_get_many)

    def before(self, obj):
        return re_Before(self, self._toregex(obj))

    def not_before(self, obj):
        return re_NotBefore(self, self._toregex(obj))

    def after(self, obj):
        return re_After(self, self._toregex(obj))

    def not_after(self, obj):
        return re_NotAfter(self, self._toregex(obj))

    def label(self, name):
        return re_Label(self, name)

    def translate(self):
        raise NotImplementedError

    def compile(self):
        return re.compile(self.translate())

    def __add__(self, other):
        return re_Cat(self, self._toregex(other))

    def __radd__(self, other):
        return re_Cat(self._toregex(other), self)

    def __or__(self, other):
        return re_Alt(self, self._toregex(other))

    def __ror__(self, other):
        return re_Alt(self._toregex(other), self)

    def __sub__(self, other):
        return re_Sub(self, self._toregex(other))

    def __rsub__(self, other):
        return re_Sub(self._toregex(other), self)

    def __neg__(self):
        return re_Neg(self)

    def __str__(self):
        return str(self.translate())

    def __getitem__(self, domain):
        return repeat(self,domain=domain)

    def selfrep(self):
        raise NotImplementedError

class re_Const(regex):
    '''
    Constant regular expression.
    '''
    def __init__(self, obj):
        if isinstance(obj, int):
            assert obj<256, "Cannot convert number to string"
            if obj<32:
                self.obj = r"\%c"+chr(obj)
            else:
                self.obj = chr(obj)
        else:
            assert isinstance(obj, basestring)
            self.obj = obj

    def translate(self):
        return self.obj

    def selfrep(self, text=True):
        out = {r"\d":"digits",
                r"\s":"white",
                r"\w":"alphanum",
                ".":"each",
                "^":"start",
                "$":"end",
                "[a-zA-Z0-9_]":"alphanum"}.get(self.obj)
        if out is None:
            if text:
                return "text('%s')"%self.obj
            else:
                return "'%s'"%self.obj
        else:
            return out

class text(re_Const):
    def __init__(self, obj):
        re_Const.__init__(self, obj)
        self._raw = obj
        self.obj  = re.escape(obj)

    def selfrep(self, text=True):
        return "text('%s')"%self._raw

class re_Neg(regex):
    def __init__(self, obj):
        if isinstance(obj,re_Const):
            if obj.obj == r"\d":
                self.obj = r"\D"
            elif obj.obj == r"\s":
                self.obj = r"\S"
            elif obj.obj == r"\b":
                self.obj = r"\B"
            if obj.obj == r"\D":
                self.obj = r"\d"
            elif obj.obj == r"\S":
                self.obj = r"\s"
            elif obj.obj == r"\B":
                self.obj = r"\b"

    def translate(self):
        return self.obj


class re_Before(regex):
    def __init__(self, obj, before):
        self.obj = obj
        self.before = before

    def translate(self):
        return "("+self.obj.translate()+"(?="+self.before.translate()+")"+")"

    def selfrep(self):
        return self.obj.selfrep()+".before(%s)"%self.before.selfrep()

class re_NotBefore(regex):
    def __init__(self, obj, before):
        self.obj = obj
        self.before = before

    def translate(self):
        return self.obj.translate()+"(?!"+self.before.translate()+")"

    def selfrep(self):
        return self.obj.selfrep()+".not_before(%s)"%self.before.selfrep()

class re_After(regex):
    def __init__(self, obj, after):
        self.obj = obj
        self.after = after

    def translate(self):
        return "(?<="+self.after.translate()+")"+self.obj.translate()

    def selfrep(self):
        return self.obj.selfrep()+".after(%s)"%self.after.selfrep()

class re_NotAfter(regex):
    def __init__(self, obj, after):
        self.obj = obj
        self.after = after

    def translate(self):
        return "(?<!"+self.after.translate()+")"+self.obj.translate()

    def selfrep(self):
        return self.obj.selfrep()+".not_after(%s)"%self.after.selfrep()

class re_Sub(regex):
    def __init__(self, obj1, obj2):
        self.obj = (obj1,obj2)

    def translate(self):
        trans1,trans2 = [obj.translate() for obj in self.obj]
        return trans1+"(?!"+trans2+")"

    def selfrep(self):
        return self.obj[0].selfrep()+" - "+self.obj[1].selfrep()


class re_Seq(regex):
    def __init__(self, obj1, obj2):
        regex.__init__(self, [])
        if isinstance(obj1, self.__class__):
            self.obj+=obj1.obj
        else:
            self.obj.append(obj1)
        if isinstance(obj2, self.__class__):
            self.obj+=obj2.obj
        else:
            self.obj.append(obj2)

class re_Cat(re_Seq):
    def translate(self):
        return "".join([obj.translate() for obj in self.obj])

    def selfrep(self):
        sreps = []
        for i,obj in enumerate(self.obj):
            if isinstance(obj, re_Const):
                if i and isinstance(self.obj[i-1], re_Const):
                    sreps[-1] = sreps[-1][:-1]+obj.selfrep(False)[1:]
                else:
                    sreps.append(obj.selfrep(False))
            else:
                sreps.append(obj.selfrep())
        return "+".join(sreps)


class re_Alt(re_Seq):
    def translate(self):
        return "("+"|".join([obj.translate() for obj in self.obj])+")"

    def selfrep(self):
        return "|".join([obj.selfrep() for obj in self.obj])

class re_Label(regex):
    labels = []
    def __init__(self, obj, label):
        self.obj = obj
        self.label = label
        self.labels.append(label)

    def translate(self):
        trans = self.obj.translate()
        if trans.startswith("(") and trans.endswith(")"):
            trans = trans[1:-1]
        return "(?P<%s>"%self.label+trans+")"

    def selfrep(self):
        return self.obj.selfrep()+".label('%s')"%self.label


class cond(regex):
    def __init__(self, label, yes_pattern, no_pattern = None):
        self.label = label
        if isinstance(yes_pattern,re_Alt):
            self.yes_pattern = yes_pattern.obj[0]
            self.no_pattern = yes_pattern.obj[1]
        else:
            self.yes_pattern = self._toregex(yes_pattern)
            if no_pattern:
                self.no_pattern  = self._toregex(no_pattern)
            else:
                self.no_pattern = None

    def translate(self):
        if self.no_pattern:
            return "(?(%s)"%self.label +self.yes_pattern.translate() + "|" + self.no_pattern.translate()
        else:
            return "(?(%s)"%self.label +self.yes_pattern.translate()

    def selfrep(self):
        if self.no_pattern:
            return "cond(%s,%s,%s)"%(self.label, self.yes_pattern.selfrep(),self.no_pattern.selfrep())
        else:
            return "cond(%s,%s)"%(self.label, self.yes_pattern.selfrep())

class match(regex):
    def __init__(self, name):
        if name not in re_Label.labels:
            raise ValueError("No label available to match against")
        self.obj = name

    def translate(self):
        return '(?P=%s)' % self.obj

    def selfrep(self):
        return "match(%s)"%self.name

class repeat(regex):
    def __init__(self, obj, domain = slice(1,None,None), greedy = True):
        regex.__init__(self, obj)
        self.slice = domain
        self._set_domain()
        if isinstance(domain, (int, long)):
            self.domain = "{%d}"%domain
        else:
            self.domain = ("{%s,%s}"%(domain.start, domain.stop)).replace("None","")
        self._greedy = greedy

    def _set_domain(self):
        if isinstance(self.slice, (int, long)):
            self.domain = "{%d}"%self.slice
        else:
            self.domain = ("{%s,%s}"%(self.slice.start, self.slice.stop)).replace("None","")

    def _get_nongreedy(self):
        self._greedy = False
        return self

    nongreedy = property(_get_nongreedy)

    def __getitem__(self, domain):
        self.domain = domain
        return self

    def translate(self):
        trans = self.obj.translate()
        if not trans.startswith("("):
            trans = "("+trans+")"
        if self._greedy:
            return trans+self.domain
        else:
            return trans+self.domain+"?"

    def selfrep(self):
        if isinstance(self.slice, (int, long)):
            return self.obj.selfrep()+"[%s]"%self.slice
        else:
            start, stop = self.slice.start, self.slice.stop
            if start is None:
                start = ""
            if stop is None:
                stop = ""
            if not self._greedy:
                greed = ".nongreedy"
            else:
                greed = ""
            s = self.obj.selfrep()
            if '+' in s or '|' in s:
                return "(%s)[%s:%s]%s"%(s,start, stop, greed)
            else:

                return "%s[%s:%s]%s"%(s, start, stop, greed)

def maybe(obj):
    return repeat(obj, domain=slice(0,1,None))

def each_but(*chars):
    return charset("^",*chars)


class charset(regex):
    def __init__(self,*args):
        self.obj = []
        for item in args:
            if isinstance(item, re_Const):
                item = item.obj
            if item == "-":
                self.obj.append(r"\-")
            elif item == "]":
                self.obj.append(r"\]")
            elif item is digits:
                self.obj.append("0-9")
            elif item is white:
                self.obj.append(" \t\n\r\f\v")
            elif isinstance(item,re_Neg):
                self.obj.append(item.obj)
            elif isinstance(item,basestring):
                self.obj.append(item)
            elif isinstance(item, charset):
                self.obj+=item.obj
            else:
                raise ValueError("Not convertible to charset %s", item)

    def translate(self):
        return "["+ "".join(self.obj)+"]"

    def selfrep(self):
        return "charset(%s)"%self.translate()[1:-1]

class adapter(regex):
    def __init__(self, initial_pattern):
        assert isinstance(initial_pattern, regex), "adapter must be initialized with regex object"
        self.pattern = initial_pattern
        self._goals = []

    def add_goals(self, *goal):
        pass

    def widen(self, **qualifiers):
        pass

    def specialize(self, **qualifiers):
        pass

    def adapt(self, **kwd):
        assert self._goals!=[], "goal set may not be empty"

def ends(obj):
    return obj+re_Const("$")

def starts(obj):
    return re_Const("^")+obj

digits      = re_Const(r'\d')
white       = re_Const(r'\s')
each        = re_Const('.')
start       = re_Const('^')
end         = re_Const('$')
alphanum    = re_Const(r'\w')
alpha       = re_Const(r'[a-zA-Z_]')
startString = re_Const(r'\A')
endString   = re_Const(r'\Z')
wordbreak   = re_Const(r'\b')

x = each+alphanum[3:].before("8989")[1]

#print x.compile().match("(abc8989i").span()

a = adapter(x)
#a.adapt()
a.widen(typ=alphanum)

if __name__ == '__main__':
    t = regex.from_re(r"^track=(.+?)")
    print t.selfrep()
