'''
NEW: general top down parser. work in progress.
'''

from eetransgen import*

class Rule(RuleTracer):
    def __init__(self, *args):
        RuleTracer.__init__(self, *args)

    def move(self, s):
        r = self.clone()
        try:
            r._selected = r.select(s)
            return r
        except NonSelectableError:
            pass

    def selection(self):
        return self._selected


class TokenStream(object):
    def __init__(self, arg):
        if isinstance(arg, list):
            self.tokstream = arg
            self.position = 0
        elif isinstance(arg, TokenStream):
            self.tokstream = arg.tokstream
            self.position = arg.position
        else:
            raise TypeError("TokenStream or list expected. %s found instead"%type(arg))
        self.size = len(self.tokstream)

    def shift(self):
        self.position+=1

    def this(self):
        if self.position<self.size:
            return self.tokstream[self.position]

class Block(object):
    No_Parse = ()
    def __init__(self, s=-1):
        self.sym = s
        self.children = []
        self.tokstream = None
        self.token = []

    def tocst(self):
        if self.token:
            return list(self.token[:3])
        elif isinstance(self.sym, str):
            return self.children[0].tocst()
        else:
            return [self.sym]+[c.tocst() for c in self.children]

class Memo(list):
    def __init__(self, memo = None):
        if memo:
            list.__init__(self, memo)
        else:
            list.__init__(self)

def memoize(f):
    def wrap(self, tokstream, rule, s, memo):
        m = Memo(memo)
        return f(self, tokstream, rule, s, m)
    return wrap


class Parser(object):
    def __init__(self, fiber = zero_fiber):
        self.fiber = fiber
        self.gt = GrammarTracer(self.fiber)

    def parse(self, tokstream):
        memo = Memo()
        return self.depth_first(tokstream, self.fiber.symbol.file_input, memo)

    @memoize
    def try_continue(self, tokstream, rule, s, memo = []):
        retBlock = self.depth_first(TokenStream(tokstream), s, memo)
        if retBlock == Block.No_Parse:
            return Block.No_Parse
        else:
            b = Block()  # empty block initialized with node id = -1
            b.children.append(retBlock)
            retChain = self.traverse_rule(TokenStream(retBlock.tokstream), rule, s, memo)
            if retChain!=Block.No_Parse:
                b.children+=retChain.children
                b.tokstream = retChain.tokstream
                return b
            else:
                return Block.No_Parse

    @memoize
    def traverse_rule(self, tokstream, rule, s0, memo = []):
        r = rule.move(s0)  # select s0 on the graph and move to the next selection
        memo.append(s0)
        if not r:
            return Block.No_Parse
        ejectable = None in r.selection()
        if ejectable:
            selection = [s for s in r.selection() if s]
        else:
            selection = r.selection()
        b = Block(s0)  # terminated
        for s in selection:
            if s not in memo:
                block = self.try_continue(tokstream, r, s, memo)
                if block is not Block.No_Parse:
                    b.tokstream = block.tokstream
                    b.children+=block.children
                    return b
        else:
            if ejectable:
                b.tokstream = tokstream
                return b
        return Block.No_Parse

    def matches(self, tok, s):
        if isinstance(s, str) and tok[0] == self.fiber.token.NAME:
            return s == tok[1]
        else:
            return s == tok[0]

    def depth_first(self, tokstream, s, memo = []):
        tok = tokstream.this()
        if self.matches(tok, s):
            ts = TokenStream(tokstream)
            ts.shift()
            b = Block(s)
            b.tokstream = ts
            b.token = tok
            return b
        r = Rule(self.gt.dfa_table)
        return self.traverse_rule(tokstream, r, s, memo)

if __name__ == '__main__':
    from EasyExtend.eecompiler import EECompiler
    import EasyExtend.fibers.new_parser.fiber as new_parser
    tokenstream = TokenStream(EECompiler(new_parser).eetokenize_string("e + e + e"))

    #print tokenstream.tokstream
    p = Parser(new_parser)
    block = p.parse(tokenstream)
    print new_parser.pprint(block.tocst())








